""" Module to handle errors in Pythran. """


class PythranInternalError(Exception):

    """ Exception raise on Incorrect internal behavior in Pythran. """


class PythranCompileError(Exception):

    """ Exception raise on when Pythran fails the compile to binary step. """


class PythranSyntaxError(SyntaxError):
    def __init__(self, msg, node=None):
        SyntaxError.__init__(self, msg)
        if node:
            self.filename = getattr(node, 'filename', None)
            self.lineno = getattr(node, 'lineno', None)
            self.offset = getattr(node, 'col_offset', None)

    def __str__(self):
        loc_info = self.lineno is not None and self.offset is not None

        if self.filename and loc_info:
            with open(self.filename) as f:
                for i in range(self.lineno - 1):
                    f.readline()  # and drop it
                extra = '{}\n{}'.format(f.readline().rstrip(),
                                        " " * (self.offset) + "^~~~ (o_0)")
        else:
            extra = None

        if loc_info:
            format_header = "{}:{}:{}"
            format_args = self.lineno, self.offset, self.args[0],
        else:
            format_header = "{}:"
            format_args = self.args[0],

        r = (format_header + " error: {}").format(
                self.filename or "<unknown>",
                *format_args)

        if extra is not None:
            r += "\n----\n"
            r += extra
            r += "\n----\n"

        return r


class PythranTypeError(PythranSyntaxError):
    "A new type to distinguish general syntax errors from typing issues"





'''
This module defines classes needed to manipulate c++ types from pythran.
'''

from inspect import isclass
from pythran.utils import cxxid


class ordered_set(object):

    def __init__(self, elements=None):
        self.values = list()
        self.unique_values = set()
        if elements is not None:
            self.extend(elements)

    def append(self, value):
        if value not in self.unique_values:
            self.values.append(value)
            self.unique_values.add(value)

    def extend(self, values):
        for value in values:
            self.append(value)

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __getitem__(self, index):
        return self.values[index]


class TypeBuilder(object):
    '''
    >>> builder = TypeBuilder()

    >>> builder.NamedType('long long')
    long long

    >>> l_ty = builder.NamedType('long')
    >>> i_ty = builder.NamedType('int')
    >>> f_ty = builder.NamedType('float')

    >>> builder.CombinedTypes(builder.NamedType('long'), builder.NamedType('char'))
    typename __combined<long,char>::type

    >>> builder.ArgumentType(4)
    std::remove_cv_t<std::remove_reference_t<argument_type4>>

    >>> builder.Assignable(builder.NamedType("long"))
    typename pythonic::assignable<long>::type

    >>> builder.Returnable(builder.NamedType("long"))
    typename pythonic::returnable<long>::type

    >>> builder.Lazy(builder.NamedType("long"))
    typename pythonic::lazy<long>::type

    >>> builder.FunctionType("toto")
    toto

    >>> builder.IteratorOfType(builder.NamedType('some'))
    typename some::iterator
    >>> builder.IteratorOfType(builder.NamedType('typename some::stuff'))
    typename some::stuff::iterator

    >>> builder.IteratorContentType(builder.NamedType('str'))
    std::remove_cv_t<typename std::iterator_traits<\
typename std::remove_reference_t<str>::iterator>::value_type>

    >>> builder.GetAttr(builder.NamedType('complex'), 'real')
    decltype(pythonic::builtins::getattr(\
pythonic::types::attr::REAL{}, std::declval<complex>()))

    >>> builder.ReturnType(builder.NamedType('math::cos'), f_ty)
    std::result_of_t<math::cos(float)>

    >>> t = builder.TupleType(i_ty, builder.NamedType('str'))
    >>> builder.ElementType(1, t)
    std::tuple_element_t<1, std::remove_reference_t<pythonic::types::make_tuple_t<int, str>>>


    >>> builder.ListType(builder.NamedType('int'))
    pythonic::types::list<std::remove_reference_t<int>>

    >>> builder.NDArrayType(builder.NamedType('int'), 1)
    pythonic::types::ndarray<int, pythonic::types::pshape<long>>

    >>> builder.SetType(builder.NamedType('int'))
    pythonic::types::set<int>

    >>> builder.TupleType(i_ty, builder.NamedType('bool'))
    pythonic::types::make_tuple_t<int, bool>

    >>> builder.DictType(builder.NamedType('int'), builder.NamedType('float'))
    pythonic::types::dict<int,float>

    >>> builder.ContainerType(builder.NamedType('int'))
    container<std::remove_reference_t<int>>

    >>> builder.IndexableType(builder.NamedType('int'))
    indexable<int>

    >>> op = lambda x,y: x + '+' + y
    >>> builder.ExpressionType(op, l_ty, i_ty)
    decltype(std::declval<long>()+std::declval<int>())
    '''

    def __init__(builder):
        builder._instances = dict()

        class Type(object):
            """
            A generic type object to be sub-classed

            The keyword arguments are used to build the internal representation
            one attribute per key with the associated value
            """

            def __new__(cls, *args, **kwargs):
                # no memoization for PType
                if cls.__name__ == 'PType':
                    return super(Type, cls).__new__(cls)

                key = cls,
                for v in args + tuple(v for k, v in sorted(kwargs.items())):
                    if isinstance(v, list):
                        v = tuple(v)
                    key += v,

                if key not in builder._instances:
                    builder._instances[key] = super(Type, cls).__new__(cls)

                return builder._instances[key]

            def __init__(self, **kwargs):
                for k, v in kwargs.items():
                    if isinstance(v, list):
                        v = tuple(v)
                    setattr(self, k, v)

            def iscombined(self):
                return False

            def sgenerate(self):
                return self.generate(lambda obj: obj.sgenerate())

            def __repr__(self):
                return self.sgenerate()  # for testing only

            def __str__(self):
                raise NotImplementedError("Call self.sgenerate() instead")

        class NamedType(Type):
            """
            A generic type object, to hold scalar types and such
            """

            def __init__(self, srepr):
                super(NamedType, self).__init__(srepr=srepr)

            def generate(self, _):
                return self.srepr

        class PType(Type):
            """
            A generic parametric type
            """

            def __init__(self, fun, ptype, index):
                super(PType, self).__init__(fun=fun,
                                            type=ptype,
                                            name=f"__ptype{index}")

            def generate(self, ctx):
                return ctx(self.type)

            def instanciate(self, caller, arguments):
                if self.fun is caller:
                    return builder.UnknownType
                else:
                    return InstantiatedType(self.fun, self.name, arguments)

        class LType(Type):
            def __init__(self, base, node):
                super(LType, self).__init__(node=node)
                self.isrec = False
                self.orig = base
                self.final_type = base

            def generate(self, ctx):
                if self.isrec:
                    return ctx(self.orig)
                else:
                    self.isrec = True
                    res = ctx(self.final_type)
                    self.isrec = False
                    return res

        class InstantiatedType(Type):
            """
            A type instantiated from a parametric type
            """
            def __init__(self, fun, name, arguments):
                super(InstantiatedType, self).__init__(fun=fun,
                                                       name=name,
                                                       arguments=arguments)

            def generate(self, ctx):
                if self.arguments:
                    args = ", ".join(ctx(arg) for arg in self.arguments)
                    template_params = f"<{args}>"
                else:
                    template_params = ""

                return f"typename {self.fun.name}::type{template_params}::{self.name}"

        class CombinedTypes(Type):
            """
            type resulting from the combination of other types

            """

            def __init__(self, *types):
                super(CombinedTypes, self).__init__(types=types)

            def iscombined(self):
                return True

            def generate(self, ctx):
                # Degenerated trees may lead to very deep call stacks.
                # In that case try hard to recover by cutting the tree
                import sys
                current_recursion_limit = sys.getrecursionlimit()
                stypes = ordered_set()
                for t in self.types:
                    try:
                        stypes.append(ctx(t))
                    except RecursionError:
                        sys.setrecursionlimit(current_recursion_limit * 2)
                        break

                if not stypes:
                    sys.setrecursionlimit(current_recursion_limit)
                    raise RecursionError
                elif len(stypes) == 1:
                    sys.setrecursionlimit(current_recursion_limit)
                    return stypes[0]
                else:
                    stmp = f'typename __combined<{",".join(stypes)}>::type'
                    sys.setrecursionlimit(current_recursion_limit)
                    return stmp

        class IntegralConstant(Type):
            """
            A generic type object, to hold scalar types and such
            """

            def __init__(self, of, index):
                super(IntegralConstant, self).__init__(of=of, index=index)

            def generate(self, ctx):
                ty = ctx(self.of)
                return f"std::integral_constant<{ty}, {str(self.index).lower()}>"

        class ArgumentType(Type):
            """
            A type to hold function arguments
            """
            def __init__(self, num):
                super(ArgumentType, self).__init__(num=num)

            def generate(self, _):
                return f'std::remove_cv_t<std::remove_reference_t<argument_type{self.num}>>'

        class DependentType(Type):
            """
            A class to be sub-classed by any type that depends on another type
            """
            def __init__(self, of):
                assert of is not None
                super(DependentType, self).__init__(of=of)

            def iscombined(self):
                return self.of.iscombined()


        class Assignable(DependentType):
            """
            A type which can be assigned

            It is used to make the difference between
            * transient types (e.g. generated from expression template)
            * assignable types (typically type of a variable)
            """

            def generate(self, ctx):
                return f'typename pythonic::assignable<{ctx(self.of)}>::type'

        class AssignableNoEscape(DependentType):
            """
            Similar to Assignable, but it doesn't escape it's declaration scope
            """

            def generate(self, ctx):
                return f'typename pythonic::assignable_noescape<{ctx(self.of)}>::type'

        class Returnable(DependentType):
            """
            A type which can be returned

            It is used to make the difference between
            * returned types (that cannot hold a reference to avoid dangling
                              reference)
            * assignable types (local to a function)

            """

            def generate(self, ctx):
                return f'typename pythonic::returnable<{ctx(self.of)}>::type'

        class Lazy(DependentType):
            """
            A type which can be a reference

            It is used to make a lazy evaluation of numpy expressions

            """

            def generate(self, ctx):
                return f'typename pythonic::lazy<{ctx(self.of)}>::type'

        class FunctionType(Type):
            """
            Gather the type of a function reference
            """

            def __init__(self, *path):
                super(FunctionType, self).__init__(path=path)

            def generate(self, _):
                return '::'.join(map(cxxid, self.path))


        class AddConst(DependentType):
            '''
            Type of an Iterator of a container
            '''
            def generate(self, ctx):
                of_type = ctx(self.of)
                return f'decltype(pythonic::types::as_const(std::declval<{of_type}>()))'

        class IteratorOfType(DependentType):
            '''
            Type of an Iterator of a container
            '''
            def generate(self, ctx):
                container_type = ctx(self.of)
                if container_type.startswith('typename'):
                    return f'{container_type}::iterator'
                else:
                    return f'typename {container_type}::iterator'

        class IteratorContentType(DependentType):
            '''
            Type of an iterator over the content of a container
            '''

            def generate(self, ctx):
                iterator_value_type = ctx(self.of)
                return f'std::remove_cv_t<typename std::iterator_traits<typename std::remove_reference_t<{iterator_value_type}>::iterator>::value_type>'

        class GetAttr(Type):
            '''
            Type of a named attribute
            '''
            def __init__(self, param, attr):
                super(GetAttr, self).__init__(param=param, attr=attr)

            def generate(self, ctx):
                attr = f'pythonic::types::attr::{self.attr.upper()}'
                return f'decltype(pythonic::builtins::getattr({attr}{{}}, std::declval<{ctx(self.param)}>()))'

        class ReturnType(Type):
            '''
            Return type of a call with arguments
            '''
            def __init__(self, ftype, *args):
                super(ReturnType, self).__init__(ftype=ftype, args=args)

            def generate(self, ctx):
                # the return type of a constructor is obvious
                cg = ctx(self.ftype)
                args = [ctx(arg) for arg in self.args]
                return f'std::result_of_t<{cg}({", ".join(args)})>'

        class ElementType(Type):
            '''
            Type of the ith element of a tuple or container
            '''

            def __init__(self, index, of):
                super(ElementType, self).__init__(of=of, index=index)

            def iscombined(self):
                return self.of.iscombined()

            def generate(self, ctx):
                return f'std::tuple_element_t<{self.index}, std::remove_reference_t<{ctx(self.of)}>>'

        class TypeType(DependentType):
            '''
            Type holding a type
            '''

            def generate(self, ctx):
                return f'pythonic::types::type_t<{ctx(self.of)}>'


        class ListType(DependentType):
            '''
            Type holding a list of stuff of the same type
            '''

            def generate(self, ctx):
                return f'pythonic::types::list<std::remove_reference_t<{ctx(self.of)}>>'

        class SetType(DependentType):
            '''
            Type holding a set of stuff of the same type
            '''

            def generate(self, ctx):
                return f'pythonic::types::set<{ctx(self.of)}>'

        class TupleType(Type):
            '''
            Type holding a tuple of stuffs of various types
            '''
            def __init__(self, *ofs):
                super(TupleType, self).__init__(ofs=ofs)

            def iscombined(self):
                return any(of.iscombined() for of in self.ofs)

            def generate(self, ctx):
                elts = (ctx(of) for of in self.ofs)
                return f'pythonic::types::make_tuple_t<{", ".join(elts)}>'

        class DictType(Type):
            '''
            Type holding a dict of stuff of the same key and value type
            '''

            def __init__(self, of_key, of_val):
                super(DictType, self).__init__(of_key=of_key, of_val=of_val)

            def iscombined(self):
                return any((of.iscombined()
                            for of in (self.of_key, self.of_val)))

            def generate(self, ctx):
                return f'pythonic::types::dict<{ctx(self.of_key)},{ctx(self.of_val)}>'

        class NDArrayType(DependentType):
            '''
            Type holding a numpy array without view
            '''
            def __init__(self, dtype, nbdims):
                super(DependentType, self).__init__(of=dtype, nbdims=nbdims)

            def generate(self, ctx):
                return f'pythonic::types::ndarray<{ctx(self.of)}, pythonic::types::pshape<{", ".join(["long"] * self.nbdims)}>>'


        class ContainerType(DependentType):
            '''
            Type of any container of stuff of the same type
            '''

            def generate(self, ctx):
                return f'container<std::remove_reference_t<{ctx(self.of)}>>'

        class IndexableType(DependentType):
            '''
            Type of any container indexed by the same type
            '''

            def generate(self, ctx):
                return f'indexable<{ctx(self.of)}>'

        class IndexableContainerType(Type):
            '''
            Type of any container of stuff of the same type,
            indexable by another type
            '''
            def __init__(self, of_key, of_val):
                super(IndexableContainerType, self).__init__(of_key=of_key,
                                                             of_val=of_val)

            def iscombined(self):
                return any((of.iscombined()
                            for of in (self.of_key, self.of_val)))

            def generate(self, ctx):
                return f'indexable_container<{ctx(self.of_key)}, std::remove_reference_t<{ctx(self.of_val)}>>'

        class ExpressionType(Type):

            """
            Result type of an operator call.
            """

            def __init__(self, op, *exprs):
                super(ExpressionType, self).__init__(op=op, exprs=exprs)

            def iscombined(self):
                return any(expr.iscombined() for expr in self.exprs)

            def generate(self, ctx):
                gexprs = [f"std::declval<{ctx(expr)}>()" for expr in self.exprs]
                return f'decltype({self.op(*gexprs)})'

        builder.UnknownType = Type()

        for objname, obj in locals().items():
            if isclass(obj):
                setattr(builder, objname, obj)



""" Globals computes the value of globals(). """

from pythran.analyses.global_declarations import GlobalDeclarations
from pythran.passmanager import ModuleAnalysis


class Globals(ModuleAnalysis[GlobalDeclarations]):

    ResultType = set

    def visit_Module(self, node):
        self.result = {'builtins',
                       '__dispatch__'}.union(self.global_declarations.keys())



""" RemoveNamedArguments turns named arguments into regular ones.  """

from pythran.analyses import Aliases, DefUseChains
from pythran.conversion import PYTHRAN_IMPORT_MANGLING
from pythran.errors import PythranInternalError
from pythran.passmanager import Transformation
from pythran.syntax import PythranSyntaxError
from pythran.tables import MODULES


import gast as ast
from copy import deepcopy


def handle_special_calls(func_alias, node):
    if func_alias is MODULES['numpy']['arange']:
        if len(node.args) == 1:
            node.args.insert(0, ast.Constant(0, None))


def same_nodes(n0, n1):
    if type(n0) is not type(n1):
        return False
    if isinstance(n0, ast.Constant):
        return type(n0.value) is type(n1.value) and n0.value == n1.value
    if isinstance(n0, ast.Attribute):
        return n0.attr == n1.attr and same_nodes(n0.value, n1.value)
    if isinstance(n0, ast.Name):
        assert n0.id.startswith(PYTHRAN_IMPORT_MANGLING), "should be an import"
        return n0.id == n1.id
    raise NotImplementedError((n0, n1))


class RemoveNamedArguments(Transformation[Aliases, DefUseChains]):
    '''
    Replace call with named arguments to regular calls

    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> code = 'def foo(x, y): return x + y\\ndef bar(z): return foo(y=z, x=0)'
    >>> node = ast.parse(code)
    >>> pm = passmanager.PassManager("test")
    >>> _, node = pm.apply(RemoveNamedArguments, node)
    >>> print(pm.dump(backend.Python, node))
    def foo(x, y):
        return (x + y)
    def bar(z):
        return foo(0, z)
    '''

    def handle_keywords(self, func, node, offset=0):
        '''
        Gather keywords to positional argument information

        Assumes the named parameter exist, raises a KeyError otherwise
        '''
        func_argument_names = {}
        for i, arg in enumerate(func.args.args[offset:]):
            assert isinstance(arg, ast.Name)
            func_argument_names[arg.id] = i

        func_argument_kwonly_names = {}
        for i, arg in enumerate(func.args.kwonlyargs):
            assert isinstance(arg, ast.Name)
            func_argument_kwonly_names[arg.id] = i

        nargs = len(func.args.args) - offset
        defaults = func.args.defaults
        keywords = {func_argument_names[kw.arg]: kw.value
                    for kw in node.keywords
                    if kw.arg not in func_argument_kwonly_names}

        keywords_only = []
        nb_kw = len(node.keywords)
        for i, kw in enumerate(list(reversed(node.keywords))):
            if kw.arg in func_argument_kwonly_names:
                keywords_only.append((func_argument_kwonly_names[kw.arg],
                                      kw.value))
                node.keywords.pop(nb_kw - i - 1)

        keywords_only = [v for _, v in sorted(keywords_only)]

        extra_keyword_offset = max(keywords.keys()) if keywords else 0
        node.args.extend([None] * (1 + extra_keyword_offset - len(node.args)))

        replacements = {}
        for index, arg in enumerate(node.args):
            if arg is None:
                if index in keywords:
                    replacements[index] = deepcopy(keywords[index])
                else:  # must be a default value
                    replacements[index] = deepcopy(defaults[index - nargs])

        if not keywords_only:
            return replacements

        node.args.append(ast.Call(
            ast.Attribute(
                ast.Attribute(
                    ast.Name("builtins", ast.Load(), None, None),
                    "pythran",
                    ast.Load()),
                "kwonly",
                ast.Load()),
            [], [])
        )
        node.args.extend(keywords_only)

        return replacements

    def visit_FunctionDef(self, node):
        self.current_function = node
        return super().generic_visit(node)

    def visit_Call(self, node):
        if node.keywords:
            self.update = True

            aliases = self.aliases[node.func]
            assert aliases, "at least one alias"

            # all aliases should have the same structural type...
            # call to self.handle_keywords raises an exception otherwise
            try:
                def visit_aliases(aliases):
                    all_replacements = []
                    for func_alias in aliases:
                        handle_special_calls(func_alias, node)

                        if func_alias is None:  # aliasing computation failed
                            pass
                        elif isinstance(func_alias, ast.Call):  # nested function
                            # func_alias looks like functools.partial(foo, a)
                            # so we reorder using alias for 'foo'
                            offset = len(func_alias.args) - 1
                            call = func_alias.args[0]
                            for func_alias in self.aliases[call]:
                                all_replacements.append(self.handle_keywords(func_alias,
                                                                    node,
                                                                             offset))
                        elif isinstance(func_alias, ast.Name):
                            if not isinstance(func_alias.ctx, ast.Param):
                                raise PythranInternalError(
                                "Unexpected non-parameter reference in call")
                            index = self.current_function.args.args.index(func_alias)
                            def_ = self.def_use_chains.chains[self.current_function]
                            for def_user in def_.users():
                                for user in def_user.users():
                                    if not isinstance(user.node, ast.Call):
                                        continue
                                    effective_arg = user.node.args[index]
                                    arg_aliases = self.aliases[effective_arg]
                                    all_replacements.append(visit_aliases(arg_aliases))


                        else:
                            all_replacements.append(self.handle_keywords(func_alias,
                                                                     node))

                    if not all_replacements:
                        return {}
                    elif len(all_replacements) == 1:
                        return all_replacements[0]
                    else:
                        replacement = {}
                        for candidate in all_replacements:
                            for k, v in candidate.items():
                                if k not in replacement:
                                    replacement[k] = v
                                else:
                                    if not same_nodes(v, replacement[k]):
                                        raise PythranSyntaxError(
                                            "different default argument values depending on the call site for this node", node)

                        return replacement

                replacements = visit_aliases(aliases)

                # if we reach this point, we should have a replacement
                # candidate, or nothing structural typing issues would have
                # raised an exception in handle_keywords
                if replacements:
                    for index, value in replacements.items():
                        node.args[index] = value
                    node.keywords = []

            except KeyError as ve:
                err = ("function uses an unknown (or unsupported) keyword "
                       "argument `{}`".format(ve.args[0]))
                raise PythranSyntaxError(err, node)
        return self.generic_visit(node)



'''
This module performs the return type inference, according to symbolic types,
   It then reorders function declarations according to the return type deps.
    * type_all generates a node -> type binding
'''

from pythran.analyses import LazynessAnalysis, StrictAliases, YieldPoints
from pythran.analyses import LocalNodeDeclarations, Immediates, RangeValues
from pythran.analyses import Ancestors
from pythran.analyses.aliases import ContainerOf
from pythran.config import cfg
from pythran.cxxtypes import TypeBuilder, ordered_set
from pythran.intrinsic import UserFunction, Class
from pythran.passmanager import ModuleAnalysis
from pythran.errors import PythranSyntaxError
from pythran.tables import operator_to_lambda, MODULES
from pythran.typing import List, Dict, Set, Tuple, NDArray, Union
from pythran.types.conversion import pytype_to_ctype, PYTYPE_TO_CTYPE_TABLE
from pythran.types.reorder import Reorder
from pythran.utils import attr_to_path, isnum, isextslice

from collections import defaultdict
from functools import reduce
import gast as ast
from itertools import islice
from copy import deepcopy
import numpy as np

alias_to_type = {
    MODULES['builtins']['int'] : int,
    MODULES['builtins']['bool'] : bool,
    MODULES['builtins']['float'] : float,
    MODULES['builtins']['str'] : str,
    MODULES['builtins']['complex'] : complex,
    MODULES['builtins']['dict'] : Dict,
    MODULES['builtins']['list'] : List,
    MODULES['builtins']['set'] : Set,
    MODULES['builtins']['tuple'] : Tuple,
    MODULES['builtins']['type'] : type,
    MODULES['builtins']['None'] : type(None),
    MODULES['numpy']['intc']: np.intc,
    MODULES['numpy']['intp']: np.intp,
    MODULES['numpy']['int64']: np.int64,
    MODULES['numpy']['int32']: np.int32,
    MODULES['numpy']['int16']: np.int16,
    MODULES['numpy']['int8']: np.int8,
    MODULES['numpy']['uintc']: np.uintc,
    MODULES['numpy']['uintp']: np.uintp,
    MODULES['numpy']['uint64']: np.uint64,
    MODULES['numpy']['uint32']: np.uint32,
    MODULES['numpy']['uint16']: np.uint16,
    MODULES['numpy']['uint8']: np.uint8,
    MODULES['numpy']['float32']: np.float32,
    MODULES['numpy']['float64']: np.float64,
    MODULES['numpy']['complex64']: np.complex64,
    MODULES['numpy']['complex128']: np.complex128,
    MODULES['numpy']['ndarray']: NDArray,
}
try:
    alias_to_type[MODULES['numpy']['float128']] = np.float128
    alias_to_type[MODULES['numpy']['complex256']] = np.complex256
except AttributeError:
    pass

def alias_key(a):
    if hasattr(a, 'path'):
        return a.path
    if hasattr(a, 'name'):
        return a.name,
    if hasattr(a, 'id'):
        return a.id,
    if isinstance(a, ast.Subscript):
        return ('subscript:',) + alias_key(a.value) + alias_key(a.slice)
    if isinstance(a, ast.Attribute):
        return ('attr:', a.attr) + alias_key(a.value)
    if isinstance(a, ast.Call):
        return ('call:',) + alias_key(a.func)
    if isinstance(a, ContainerOf):
        return sum((alias_key(c) for c in sorted(a.containees, key=alias_key)), (str(a.index),))
    if isinstance(a, ast.Constant):
        return ('cst:', str(a.value))
    # FIXME: how could we order those?
    return str(id(a)),



class TypeAnnotationParser(ast.NodeVisitor):

    class TypeOf:
        def __init__(self, val):
            self.val = val

    class UnionOf:
        def __init__(self, lhs, rhs):
            self.lhs = lhs
            self.rhs = rhs

    def __init__(self, type_visitor):
        self.type_visitor = type_visitor
        self.aliases = self.type_visitor.strict_aliases

    def extract(self, node):
        node_aliases = self.aliases[node]
        if len(node_aliases) > 1:
            raise PythranSyntaxError("Ambiguous identifier in type annotation",
                                     node)
        if not node_aliases:
            raise PythranSyntaxError("Unbound identifier in type annotation",
                                     node)
        node_alias, = node_aliases
        if node_alias not in alias_to_type:
            raise PythranSyntaxError("Unsupported identifier in type annotation",
                                     node)

        return alias_to_type[node_alias]


    def visit_Attribute(self, node):
        return self.extract(node)

    def visit_Name(self, node):
        return self.extract(node)

    def visit_Constant(self, node):
        return node.value

    def visit_Call(self, node):
        func = self.visit(node.func)
        if func is not type:
            raise PythranSyntaxError("Expecting a type or a call to `type(...)`",
                                     node)
        if len(node.args) != 1:
            raise PythranSyntaxError("`type` only supports a single argument",
                                     node)
        self.type_visitor.visit(node.args[0])
        return self.TypeOf(self.type_visitor.result[node.args[0]])

    def visit_Tuple(self, node):
        return tuple([self.visit(elt) for elt in node.elts])

    def visit_BinOp(self, node):
        if not isinstance(node.op, ast.BitOr):
            raise PythranSyntaxError("Unsupported operation between type operands",
                                     node)
        left = self.visit(node.left)
        right = self.visit(node.right)
        return self.UnionOf(left, right)

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        slice_ = self.visit(node.slice)
        if issubclass(value, NDArray):
            dtype, ndims = slice_
            return value[tuple([dtype, *([slice(0)] * ndims)])]
        else:
            return value[slice_]


def build_type(builder, t):
    """ Python -> pythonic type binding. """
    if t is None:
        return builder.NamedType(pytype_to_ctype(type(None)))
    elif isinstance(t, List):
        return builder.ListType(build_type(builder, t.__args__[0]))
    elif isinstance(t, Set):
        return builder.SetType(build_type(builder, t.__args__[0]))
    elif isinstance(t, Dict):
        tkey, tvalue = t.__args__
        return builder.DictType(build_type(builder, tkey), build_type(builder,
                                                                      tvalue))
    elif isinstance(t, Tuple):
        return builder.TupleType(*[build_type(builder, p) for p in t.__args__])
    elif isinstance(t, NDArray):
        return builder.NDArrayType(build_type(builder, t.__args__[0]), len(t.__args__) - 1)
    elif isinstance(t, TypeAnnotationParser.TypeOf):
        return t.val
    elif isinstance(t, TypeAnnotationParser.UnionOf):
        return builder.CombinedTypes(build_type(builder, t.lhs),
                                     build_type(builder, t.rhs))
    elif t in PYTYPE_TO_CTYPE_TABLE:
        return builder.NamedType(PYTYPE_TO_CTYPE_TABLE[t])
    else:
        raise NotImplementedError("build_type on {}".format(type(t)))


def parse_type_annotation(type_visitor, ann):
    tap = TypeAnnotationParser(type_visitor)
    typ = tap.visit(ann)
    return build_type(type_visitor.builder, typ)


class UnboundableRValue(Exception):
    pass


class Types(ModuleAnalysis[Reorder, StrictAliases, LazynessAnalysis,
                           Immediates, RangeValues, Ancestors]):

    """ Infer symbolic type for all AST node. """
    class ResultType(dict):
        def __init__(self):
            self.builder = TypeBuilder()

        def copy(self):
            other = TypeResult()
            other.update(self.items())
            other.builder = self.builder
            return other


    def __init__(self):
        super().__init__()
        self.max_seq_size = cfg.getint('typing',
                                       'max_heterogeneous_sequence_size')
        self.builder = self.result.builder
        self.result["bool"] = self.builder.NamedType("bool")
        self.combiners = defaultdict(UserFunction)
        self.current_global_declarations = dict()
        self.max_recompute = 1  # max number of use to be lazy
        self.curr_locals_declaration = None
        self.ptype_count = 0

    def combined(self, *types):
        all_types = ordered_set()
        for ty in types:
            if isinstance(ty, self.builder.CombinedTypes):
                all_types.extend(ty.types)
            elif ty is not self.builder.UnknownType:
                all_types.append(ty)

        if len(all_types) == 0:
            return self.builder.UnknownType
        elif len(all_types) == 1:
            return next(iter(all_types))
        else:
            all_types = all_types[:cfg.getint('typing', 'max_combiner')]
            if {type(ty) for ty in all_types} == {self.builder.ListType}:
                return self.builder.ListType(self.combined(*[ty.of for ty in all_types]))
            if {type(ty) for ty in all_types} == {self.builder.SetType}:
                return self.builder.SetType(self.combined(*[ty.of for ty in all_types]))
            if {type(ty) for ty in all_types} == {self.builder.Assignable}:
                return self.builder.Assignable(self.combined(*[ty.of for ty in all_types]))
            if {type(ty) for ty in all_types} == {self.builder.Lazy}:
                return self.builder.Lazy(self.combined(*[ty.of for ty in all_types]))

            return self.builder.CombinedTypes(*all_types)



    def prepare(self, node):
        """
        Initialise values to prepare typing computation.

        Reorder functions to avoid dependencies issues and prepare typing
        computation setting typing values for Pythonic functions.
        """

        def register(name, module):
            """ Recursively save function typing and combiners for Pythonic."""
            for fname, function in module.items():
                if isinstance(function, dict):
                    register(name + "::" + fname, function)
                else:
                    tname = 'pythonic::{0}::functor::{1}'.format(name, fname)
                    self.result[function] = self.builder.NamedType(tname)
                    self.combiners[function] = function
                    if isinstance(function, Class):
                        register(name + "::" + fname, function.fields)

        for mname, module in MODULES.items():
            register(mname, module)
        super(Types, self).prepare(node)

    def visit_Module(self, node):
        self.generic_visit(node)
        for head in self.current_global_declarations.values():
            if head not in self.result:
                self.result[head] = "pythonic::types::none_type"

    def register(self, fname, nid, ptype):
        """register ptype as a local typedef"""
        # Too many of them leads to memory burst
        if len(self.typedefs[fname, nid]) < cfg.getint('typing',
                                                       'max_interprocedural_combiner'):
            self.typedefs[fname, nid].append(ptype)
            return True
        return False

    def node_to_id(self, n, depth=()):
        name, depth = self.node_to_name(n, depth)
        return name.id, depth

    def node_to_name(self, n, depth=()):
        if isinstance(n, ast.Name):
            return (n, depth)
        elif isinstance(n, ast.Subscript):
            if isinstance(n.slice, ast.Slice):
                return self.node_to_name(n.value, depth)
            else:
                index = n.slice.value if isnum(n.slice) else None
                return self.node_to_name(n.value, depth + (index,))
        # use alias information if any
        elif isinstance(n, ast.Call):
            for alias in self.sorted_strict_aliases(n):
                if alias is n:  # no specific alias info
                    continue
                try:
                    return self.node_to_name(alias, depth)
                except UnboundableRValue:
                    continue
        raise UnboundableRValue()

    def isargument(self, node):
        """ checks whether node aliases to a parameter."""
        try:
            node_id, _ = self.node_to_id(node)
            return (node_id in self.name_to_nodes and
                    any(isinstance(n, ast.Name) and
                        isinstance(n.ctx, ast.Param)
                        for n in self.name_to_nodes[node_id]))
        except UnboundableRValue:
            return False

    def combine(self, node, op, othernode):
        """
        Change `node` typing with combination of `node` and `othernode`.
        """
        if self.result[othernode] is self.builder.UnknownType:
            if node not in self.result:
                self.result[node] = self.builder.UnknownType
            return

        if op is None:
            op = lambda x: x

        node_aliases = ordered_set([node])
        if node in self.strict_aliases:
            node_aliases.extend(self.sorted_strict_aliases(node))
        for a in node_aliases:
            self.combine_(a, op, othernode)

    def combine_(self, node, op, othernode):
        # This comes from an assignment,so we must check where the value is
        # assigned
        name = None
        try:
            name, depth = self.node_to_name(node)
            if depth:
                former_op = op

                # update the type to reflect container nesting
                def merge_container_type(ty, index):
                    # integral index make it possible to correctly
                    # update tuple type
                    if isinstance(index, int):
                        kty = self.builder.IntegralConstant(
                                self.builder.NamedType("long"),
                                index)
                        return self.builder.IndexableContainerType(kty,
                                                                   ty)
                    elif isinstance(index, float):
                        kty = self.builder.NamedType('double')
                        return self.builder.IndexableContainerType(kty,
                                                                   ty)
                    else:
                        # FIXME: what about other key types?
                        return self.builder.ContainerType(ty)

                for node_alias in self.sorted_strict_aliases(name,
                                                             extra=[name]):
                    def traverse_alias(alias, depth, l):
                        # not interested in aliases too deep
                        if len(depth) <= l:
                            return
                        if isinstance(alias, ContainerOf):
                            d = depth[l]
                            if d is None or np.isnan(alias.index) or d == alias.index:
                                for containee in sorted(alias.containees, key=alias_key):
                                    traverse_alias(containee, depth, l + 1)
                        else:
                            def local_op(*args):
                                t = reduce(merge_container_type,
                                              depth[l:],
                                              former_op(*args))
                                return t
                            self.combine_(alias, local_op, othernode)

                    traverse_alias(node_alias, depth, 0)
                return
        except UnboundableRValue:
            pass

        if isinstance(node, ContainerOf):
            def containeeop(*args):
                container_type = op(*args)
                if isinstance(container_type, self.builder.IndexableType):
                    raise NotImplementedError
                if isinstance(container_type, (self.builder.ListType,
                                               self.builder.SetType)):
                    return container_type.of
                if isinstance(container_type, self.builder.IndexableContainerType):
                    if np.isnan(node.index):
                        return container_type.of_val
                    elif isinstance(container_type.of_key, self.builder.IntegralConstant):
                        if container_type.of_key.index != node.index:
                            raise NotImplementedError
                return self.builder.ElementType(
                        0 if np.isnan(node.index) else node.index,
                        container_type)

            for containee in sorted(node.containees, key=alias_key):
                try:
                    self.combine(containee, containeeop, othernode)
                except NotImplementedError:
                    pass

        # perform inter procedural combination
        if self.isargument(node):
            node_id, _ = self.node_to_id(node)
            if node not in self.result:
                self.result[node] = op(self.result[othernode])
            self.name_to_nodes[name.id].append(node)
            assert self.result[node], "found an alias with a type"

            parametric_type = self.builder.PType(self.current,
                                                 self.result[othernode],
                                                 self.ptype_count)
            self.ptype_count += 1

            if self.register(self.current, node_id, parametric_type):

                current_function = self.combiners[self.current]

                def translator_generator(args, op):
                    ''' capture args for translator generation'''
                    def interprocedural_type_translator(s, n):
                        translated_othernode = ast.Name(
                            '__fake__', ast.Load(), None, None)
                        s.result[translated_othernode] = (
                            parametric_type.instanciate(
                                s.current,
                                [s.result[arg] for arg in n.args]))

                        # look for modified argument
                        for p, effective_arg in enumerate(n.args):
                            formal_arg = args[p]
                            if formal_arg.id == node_id:
                                translated_node = effective_arg
                                break
                        try:
                            s.combine(translated_node,
                                      op,
                                      translated_othernode)
                        except NotImplementedError:
                            pass
                            # this may fail when the effective
                            # parameter is an expression
                        except UnboundLocalError:
                            pass
                            # this may fail when translated_node
                            # is a default parameter
                    return interprocedural_type_translator

                translator = translator_generator(self.current.args.args, op)
                current_function.add_combiner(translator)
        else:
            self.update_type(node, op, self.result[othernode])
            if name is not None:
                self.name_to_nodes[name.id].append(node)

    def update_type(self, node, ty_builder, *args):
        if ty_builder is None:
            ty, = args
        # propagate UnknowType status if one of *args is itself unknown.
        elif any(arg is self.builder.UnknownType for arg in args):
            ty = self.builder.UnknownType
        else:
            ty = ty_builder(*args)

        curr_ty = self.result.get(node, self.builder.UnknownType)
        if isinstance(curr_ty, tuple):
            return

        self.result[node] = self.combined(curr_ty, ty)

    def visit_FunctionDef(self, node):
        self.delayed_nodes = set()
        self.curr_locals_declaration = self.gather(
            LocalNodeDeclarations,
            node)
        self.current = node
        self.typedefs = defaultdict(list)
        self.name_to_nodes = defaultdict(ordered_set)
        for arg in node.args.args:
            self.name_to_nodes[arg.id].append(arg)

        self.yield_points = self.gather(YieldPoints, node)

        self.generic_visit(node)

        for delayed_node in self.delayed_nodes:
            delayed_type = self.result[delayed_node]
            if not isinstance(delayed_type, self.builder.LType):
                continue
            all_types = ordered_set(self.result[ty] for ty in
                                    self.name_to_nodes[delayed_node.id])
            final_type = self.combined(*all_types)
            delayed_type.final_type = final_type
            if final_type is delayed_type.orig:
                self.result[delayed_node] = delayed_type.orig

        # propagate type information through all aliases
        for name, nodes in self.name_to_nodes.items():
            all_types = ordered_set(self.result[ty] for ty in nodes)
            final_type = self.combined(*all_types)
            for n in nodes:
                if n not in self.delayed_nodes:
                    self.result[n] = final_type
        self.current_global_declarations[node.name] = node

        # return type may be unset if the function always raises
        return_type = self.result.get(
            node,
            self.builder.NamedType("pythonic::types::none_type"))

        self.result[node] = (
            self.builder.Returnable(return_type),
            reduce(list.__add__, self.typedefs.values(), []))
        for k in self.curr_locals_declaration:
            self.result[k] = self.get_qualifier(k)(self.result[k])

    def assignable(self, ty):
        if isinstance(ty, (self.builder.Assignable, self.builder.ListType,
                           self.builder.NamedType)):
            return ty
        else:
            return self.builder.Assignable(ty)

    def lazy(self, ty):
        if isinstance(ty, (self.builder.Lazy, self.builder.ListType,
                           self.builder.NamedType)):
            return ty
        else:
            return self.builder.Lazy(ty)

    def get_qualifier(self, node):
        lazy_res = self.lazyness_analysis[node.id]
        return (self.lazy
                if lazy_res <= self.max_recompute
                else self.assignable)

    def visit_Return(self, node):
        """ Compute return type and merges with others possible return type."""
        self.generic_visit(node)
        # No merge are done if the function is a generator.
        if not self.yield_points:
            assert node.value, "Values were added in each return statement."
            self.update_type(self.current, None, self.result[node.value])

    def visit_Yield(self, node):
        """ Compute yield type and merges it with others yield type. """
        self.generic_visit(node)
        self.update_type(self.current, None, self.result[node.value])

    def visit_Assign(self, node):
        self.visit(node.value)
        for t in node.targets:
            self.combine(t, None, node.value)
            if t in self.curr_locals_declaration:
                self.result[t] = self.get_qualifier(t)(self.result[t])
            if isinstance(t, ast.Subscript):
                self.visit_AssignedSubscript(t)

    def visit_AnnAssign(self, node):
        node_type = parse_type_annotation(self, node.annotation)
        t = node.target
        if node_type:
            self.result[t] = node_type
        if not node.value:
            self.curr_locals_declaration.remove(t)
            return
        self.visit(node.value)
        if node_type:
            # A bit odd, isn't it? :-)
            self.combine(t, None, t)
        else:
            self.combine(t, None, node.value)
        if t in self.curr_locals_declaration:
            self.result[t] = self.get_qualifier(t)(self.result[t])

        if isinstance(t, ast.Subscript):
            self.visit_AssignedSubscript(t)

    def visit_AugAssign(self, node):
        # No visit_AssignedSubscript as the container should already have been
        # populated.
        if isinstance(node.target, ast.Subscript):
            self.visit(node.target)
            self.visit(node.value)
        else:
            tmp = ast.BinOp(deepcopy(node.target), node.op, node.value)
            self.visit(tmp)
            self.combine(node.target, None, tmp)


    def visit_For(self, node):
        self.visit(node.iter)
        self.combine(node.target, self.builder.IteratorContentType, node.iter)
        for n in node.body + node.orelse:
            self.visit(n)

    def visit_BoolOp(self, node):
        """
        Merge BoolOp operand type.

        BoolOp are "and" and "or" and may return any of these results so all
        operands should have the combinable type.
        """
        # Visit subnodes
        self.generic_visit(node)
        # Merge all operands types.
        for value in node.values:
            self.update_type(node, None, self.result[value])

    def visit_BinOp(self, node):
        self.generic_visit(node)
        self.update_type(node,
                         self.builder.ExpressionType,
                         operator_to_lambda[type(node.op)],
                         self.result[node.left],
                         self.result[node.right])

    def visit_UnaryOp(self, node):
        self.generic_visit(node)
        self.update_type(node,
                         self.builder.ExpressionType,
                         operator_to_lambda[type(node.op)],
                         self.result[node.operand])

    def visit_IfExp(self, node):
        self.generic_visit(node)
        self.update_type(node, None, self.result[node.body])
        self.update_type(node, None, self.result[node.orelse])

    def visit_Compare(self, node):
        self.generic_visit(node)
        all_compare = zip(node.ops,
                          [node.left] + node.comparators[:-1],
                          node.comparators)

        for op, left, right in all_compare:
            self.update_type(node,
                             self.builder.ExpressionType,
                             operator_to_lambda[type(op)],
                             self.result[left],
                             self.result[right])

    def sorted_strict_aliases(self, func, extra=[]):
        return sorted(self.strict_aliases[func] | set(extra), key=alias_key)

    def visit_Call(self, node):
        self.generic_visit(node)

        func = node.func

        for alias in self.sorted_strict_aliases(func):
            # this comes from a bind
            if isinstance(alias, ast.Call):
                a0 = alias.args[0]
                # by construction of the bind construct
                assert len(self.strict_aliases[a0]) == 1
                bounded_function = next(iter(self.sorted_strict_aliases(a0)))
                fake_name = deepcopy(a0)
                fake_node = ast.Call(fake_name, alias.args[1:] + node.args,
                                     [])
                self.combiners[bounded_function].combiner(self, fake_node)

            # handle backward type dependencies from function calls
            else:
                self.combiners[alias].combiner(self, node)

        UnknownType = self.builder.UnknownType

        # recurring nightmare
        def last_chance():
            # maybe we can get saved if we have a hint about
            # the called function return type
            for alias in self.sorted_strict_aliases(func):
                if alias is self.current and alias in self.result:
                    # great we have a (partial) type information
                    self.result[node] = self.result[alias]
                    return
            self.result[node] = UnknownType

        if self.result[node.func] is UnknownType:
            return last_chance()

        if any(self.result[arg] is UnknownType for arg in node.args):
            return last_chance()

        # special handler for getattr: use the attr name as an enum member
        if (isinstance(func, ast.Attribute) and func.attr == 'getattr'):
            self.update_type(node,
                             self.builder.GetAttr,
                             self.result[node.args[0]],
                             node.args[1].value)
        # default behavior
        else:
            self.update_type(node,
                             self.builder.ReturnType,
                             self.result[func],
                             *[self.result[arg] for arg in node.args])

    def visit_Constant(self, node):
        """ Set the pythonic constant type. """
        ty = type(node.value)
        if ty is str and len(node.value) == 1:
            sty = 'pythonic::types::chr'
        else:
            sty = pytype_to_ctype(ty)
        if node in self.immediates:
            if sty == 'pythonic::types::chr':
                sty = self.builder.IntegralConstant(
                        self.builder.NamedType("char"),
                        f"'{node.value}'")
            else:
                sty = self.builder.IntegralConstant(
                        self.builder.NamedType(sty),
                        node.value)
        else:
            sty = self.builder.NamedType(sty)
        self.result[node] = sty

    def visit_Attribute(self, node):
        """ Compute typing for an attribute node. """
        obj, path = attr_to_path(node)
        # If no type is given, use a decltype
        if obj.isliteral():
            typename = pytype_to_ctype(obj.signature)
            self.result[node] = self.builder.NamedType(typename)
        else:
            self.result[node] = self.builder.FunctionType(*path)

    def visit_Slice(self, node):
        """
        Set slicing type using continuous information if provided.

        Also visit subnodes as they may contains relevant typing information.
        """
        self.generic_visit(node)
        nstep = node.step
        if nstep is None or (isnum(nstep) and nstep.value > 0):
            if nstep is None or nstep.value == 1:
                if all(self.range_values[p].low >= 0
                       for p in (node.lower, node.upper)):
                    ntype = "pythonic::types::fast_contiguous_slice"
                else:
                    ntype = "pythonic::types::contiguous_slice"
            else:
                ntype = "pythonic::types::cstride_slice<{}>".format(nstep.value)
            self.result[node] = self.builder.NamedType(ntype)
        else:
            self.result[node] = self.builder.NamedType(
                'pythonic::types::slice')

    def stores_to(self, node):
        ancestors = self.ancestors[node] + (node,)
        stmt_indices = [i for i, n in enumerate(ancestors)
                        if isinstance(n, (ast.Assign, ast.For))]
        if not stmt_indices:
            return True

        stmt_index = stmt_indices[-1]

        if isinstance(ancestors[stmt_index], ast.Assign):
            return ancestors[stmt_index + 1] is ancestors[stmt_index].value
        else:
            return ancestors[stmt_index + 1] is not ancestors[stmt_index].target

    def visit_Subscript(self, node):
        self.visit(node.value)
        if self.stores_to(node):
            self.result[node.value] = self.builder.AddConst(self.result[node.value])
        # type of a[1:2, 3, 4:1] is the type of: declval(a)(slice, long, slice)
        if isextslice(node.slice):
            self.visit(node.slice)
            self.update_type(node,
                             self.builder.ExpressionType,
                             lambda a, *b: "{0}({1})".format(a, ", ".join(b)),
                             self.result[node.value],
                             *[self.result[d] for d in node.slice.elts])
        elif isnum(node.slice) and node.slice.value >= 0:
            # type of a[2] is the type of an elements of a
            # this special case is to make type inference easier
            # for the back end compiler
            self.update_type(node,
                             self.builder.ElementType,
                             node.slice.value,
                             self.result[node.value])
        else:
            # type of a[i] is the return type of the matching function
            self.visit(node.slice)
            self.update_type(node,
                             self.builder.ExpressionType,
                             "{0}[{1}]".format,
                             self.result[node.value],
                             self.result[node.slice])

    def visit_AssignedSubscript(self, node):
        if isinstance(node.slice, ast.Slice):
            return
        elif isextslice(node.slice):
            return
        else:
            self.visit(node.slice)
            self.combine(node.value, self.builder.IndexableType, node.slice)

    def delayed(self, node):
        fallback_type = self.combined(*[self.result[n] for n in
                                        self.name_to_nodes[node.id]])
        self.delayed_nodes.add(node)
        return self.builder.LType(fallback_type, node)

    def visit_Name(self, node):
        if node.id in self.name_to_nodes:
            self.result[node] = self.delayed(node)
        elif node.id in self.current_global_declarations:
            newtype = self.builder.NamedType(
                self.current_global_declarations[node.id].name)
            if node not in self.result:
                self.result[node] = newtype
        else:
            self.result[node] = self.builder.UnknownType

    def visit_List(self, node):
        """ Define list type from all elements type (or empty_list type). """
        self.generic_visit(node)
        if node.elts:
            for elt in node.elts[:self.max_seq_size]:
                self.update_type(node, self.builder.ListType, self.result[elt])
        else:
            self.update_type(node,
                             self.builder.NamedType,
                             "pythonic::types::empty_list")

    def visit_Set(self, node):
        """ Define set type from all elements type (or empty_set type). """
        self.generic_visit(node)
        if node.elts:
            for elt in node.elts[:self.max_seq_size]:
                self.update_type(node, self.builder.SetType, self.result[elt])
        else:
            self.update_type(node, self.builder.NamedType,
                "pythonic::types::empty_set")

    def visit_Dict(self, node):
        """ Define set type from all elements type (or empty_dict type). """
        self.generic_visit(node)
        if node.keys:
            for key, value in islice(zip(node.keys, node.values),
                                     self.max_seq_size):
                self.update_type(node,
                                 self.builder.DictType,
                                 self.result[key],
                                 self.result[value])
        else:
            self.update_type(node,
                             self.builder.NamedType,
                             "pythonic::types::empty_dict")

    def visit_ExceptHandler(self, node):
        if node.type and node.name:
            if not isinstance(node.type, ast.Tuple):
                tname = self.builder.NamedType(
                    'pythonic::types::{0}'.format(node.type.attr))
                self.result[node.type] = tname
                self.combine(node.name, None, node.type)
        for n in node.body:
            self.visit(n)

    def visit_Tuple(self, node):
        self.generic_visit(node)
        types = [self.result[elt] for elt in node.elts]
        self.update_type(node,
                         self.builder.TupleType,
                         *types)

    def visit_arguments(self, node):
        for i, arg in enumerate(node.args):
            self.update_type(arg, self.builder.ArgumentType, i)
        for n in node.defaults:
            self.visit(n)
